#
# This patch modifies the Ruby File methods that are typically used to
# check permissions of a file (readable?, writable?, executable?, etc).
# When this patch is loaded on Mac OS X, the methods will then work
# with ACLs in addition to POSIX permissions.
#
# Copyright (c) 2010 Brian D. Wells
#
# This patch is free software: you can redistribute it and/or modify it 
# under the terms of the license of Ruby. http://www.ruby-lang.org/en/LICENSE.txt
#
# This patch is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You can contact the author by email at <spam_brian@me.com> or you can
# view his blog entry about the patch:
#
# http://www.briandwells.com/main/Blog/Entries/2010/9/30_Patch_Ruby_File.readable_to_check_ACLs.html
#

require 'inline'

module DarwinACLFilePatch
  def self.included(base)
    base.extend(ClassMethods)
    base.send(:include, InstanceMethods)
    base.class_eval do
      class << self
        alias_method :readable_without_dacl?, :readable? unless method_defined?(:readable_without_dacl?)
        alias_method :readable?, :readable_with_dacl?
        alias_method :readable_real_without_dacl?, :readable_real? unless method_defined?(:readable_real_without_dacl?)
        alias_method :readable_real?, :readable_real_with_dacl?
        alias_method :writable_without_dacl?, :writable? unless method_defined?(:writable_without_dacl?)
        alias_method :writable?, :writable_with_dacl?
        alias_method :writable_real_without_dacl?, :writable_real? unless method_defined?(:writable_real_without_dacl?)
        alias_method :writable_real?, :writable_real_with_dacl?
        alias_method :executable_without_dacl?, :executable? unless method_defined?(:executable_without_dacl?)
        alias_method :executable?, :executable_with_dacl?
        alias_method :executable_real_without_dacl?, :executable_real? unless method_defined?(:executable_real_without_dacl?)
        alias_method :executable_real?, :executable_real_with_dacl?
      end
    end
  end
  module ClassMethods
  
    def readable_with_dacl?(file_name)
	  result = check_user_access_for_file(file_name,Process.euid,ACL_READ_DATA)
	  if result == ACL_UNDEFINED_TAG
		readable_without_dacl?(file_name)
	  else
	    (result == ACL_EXTENDED_ALLOW)
	  end
    end
    def readable_real_with_dacl?(file_name)
      result = check_user_access_for_file(file_name,Process.uid,ACL_READ_DATA)
	  if result == ACL_UNDEFINED_TAG
		readable_real_without_dacl?(file_name)
	  else
	    (result == ACL_EXTENDED_ALLOW)
	  end
    end
    def writable_with_dacl?(file_name)
      result = check_user_access_for_file(file_name,Process.euid,ACL_WRITE_DATA)
	  if result == ACL_UNDEFINED_TAG
		writable_without_dacl?(file_name)
	  else
	    (result == ACL_EXTENDED_ALLOW)
	  end
    end
    def writable_real_with_dacl?(file_name)
      result = check_user_access_for_file(file_name,Process.uid,ACL_WRITE_DATA)
	  if result == ACL_UNDEFINED_TAG
		writable_real_without_dacl?(file_name)
	  else
	    (result == ACL_EXTENDED_ALLOW)
	  end
    end
    def executable_with_dacl?(file_name)
      result = check_user_access_for_file(file_name,Process.euid,ACL_EXECUTE)
	  if result == ACL_UNDEFINED_TAG
		executable_without_dacl?(file_name)
	  else
	    (result == ACL_EXTENDED_ALLOW)
	  end
    end
    def executable_real_with_dacl?(file_name)
      result = check_user_access_for_file(file_name,Process.uid,ACL_EXECUTE)
	  if result == ACL_UNDEFINED_TAG
		executable_real_without_dacl?(file_name)
	  else
	    (result == ACL_EXTENDED_ALLOW)
	  end
    end

	private

	ACL_READ_DATA = 2
	ACL_WRITE_DATA = 4
	ACL_EXECUTE = 8
	ACL_UNDEFINED_TAG = 0
	ACL_EXTENDED_ALLOW = 1
	ACL_EXTENDED_DENY = 2
  
    inline(:C) do |builder|
	  builder.include '<sys/types.h>'
	  builder.include '<sys/acl.h>'
	  builder.include '<membership.h>'
      builder.alias_type_converter 'unsigned int', 'uid_t'
      builder.alias_type_converter 'unsigned int', 'acl_tag_t'
      builder.alias_type_converter 'unsigned int', 'acl_perm_t'
	  builder.c <<EOF
		acl_tag_t check_user_access_for_file(const char *file_name, uid_t userid, acl_perm_t perm) {
			acl_tag_t result = ACL_UNDEFINED_TAG;
			uuid_t user_uuid;
			acl_t file_acl;
			acl_entry_t acl_entry;
			acl_tag_t acl_tag;
			acl_permset_t acl_permset;
			uuid_t *acl_qual;
			int er;
			int is_member;
						
			if (perm == 0)
				return result;
				
			if (mbr_uid_to_uuid(userid, user_uuid) == 0) {
				file_acl = acl_get_file(file_name, ACL_TYPE_EXTENDED);
				if (file_acl != (acl_t)NULL) {
					
					for (er = acl_get_entry(file_acl, ACL_FIRST_ENTRY, &acl_entry);
						 er == 0 && result == ACL_UNDEFINED_TAG;
						 er = acl_get_entry(file_acl, ACL_NEXT_ENTRY, &acl_entry)) {
				
						if (acl_get_tag_type(acl_entry, &acl_tag) == 0 && 
							(acl_tag == ACL_EXTENDED_ALLOW || acl_tag == ACL_EXTENDED_DENY)) {
							if (acl_get_permset(acl_entry, &acl_permset) == 0) {
								if (acl_get_perm_np(acl_permset, perm) == 1) {
									// assuming only one permission specified in perm
									acl_qual = (uuid_t *)acl_get_qualifier(acl_entry);
									if (acl_qual != NULL) {
										if (memcmp(*acl_qual, user_uuid, sizeof(user_uuid)) == 0) {
											result = acl_tag;
										} else {
											if (mbr_check_membership(user_uuid, *acl_qual, &is_member) == 0) {
												if (is_member) {
													result = acl_tag;
												}
											}
										}
										acl_free(acl_qual);
									}
								}
							}
						}
					}
					acl_free(file_acl);
				}
			}
			return result;
		}
EOF
	end
   
	private :check_user_access_for_file
	 
  end
  module InstanceMethods
  end
end
File.send(:include,DarwinACLFilePatch)
